#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <pwd.h>

#include "version.h"

static char SccsId[] = "@(#)approve.c	1.9\t7/6/93" ;

/*
 *	Copyright (c) 1993 by California Institute of Technology.
 *	Written by William Deich.  Not derived from licensed software.

 *	Permission is granted to anyone to use this software for any
 *	purpose on any computer system, and to redistribute it freely,
 *	subject to the following restrictions:
 *
 *	1. The author is not responsible for the consequences of use of this
 *	   software, no matter how awful, even if they arise from defects in it.
 *
 *	2. The origin of this software must not be misrepresented, either
 *	   by explicit claim or by omission.
 *
 *	3. Altered versions must be plainly marked as such, and must not
 *	   be misrepresented as being the original software.
 */


static FILE *fp;

extern errno;

#define MIN(a,b) ((a) < (b) ? (a) : (b))
#define UNTIL(e) while(!(e))


extern char *user;
extern char *prog;

/* approve() returns:
 *	- NULL ptr if error:
 *		a) username not found;
 *		b) superfile can't be opened for reading;
 *		c) no such command as usrcmd in superfile;
 *		d) user not allowed to execute this command.
 *	- ptr to empty string if all ok, but no program should be executed.
 *	- ptr to path of file to exec, if user allowed to do so.
 *	Any error also generates a message to stderr.

 * New calls to approve() overwrite the buffer containing the returned path.
 */

struct Ebuf {		/* an expandable buffer */
    char *buf;
    int l;
    int nalloc;
};
static struct Ebuf ebuf = { NULL, 0, 0 };

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* For printing the errors uncovered in matching user/group/host patterns */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
char *
paterror(errmsg, origtext, line, nl, file)
char *errmsg;
char *origtext;	/* the original text with the problem pattern */
int line, nl;	/* the first line + number of lines in the input block */
char *file;	/* the input file */
{
    if (nl <= 1)
	(void) fprintf(stderr, "%s: %s.  Line %d in file `%s'\n",
			    prog, errmsg, line, file);
    else
	(void) fprintf(stderr, "%s: %s.  Lines %d..%d in file `%s'\n",
			    prog, errmsg, line, line+nl-1, file);
    (void) fprintf(stderr, "\tOffending text: %s\n", origtext);
    return errmsg;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* Copies in to out, prefixing with "^" and suffixing with "$"
 * if these are missing.
 */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
void
anchor(in, out)
char *in;
char *out;
{
    int i;
    i = (*in != '^');
    if (i)
	out[0] = '^';
    (void) strcpy(out+i, in);
    i = strlen(out);
    if (out[i-1] != '$')
	out[i++] = '$';
    out[i] = '\0';
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* Grow an expandable buffer by 2K */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
char *
grow(cb)
struct Ebuf *cb;
{
    char *realloc(), *malloc();
    if (cb->buf)
	cb->buf = realloc(cb->buf, cb->nalloc += 2048);
    else
	cb->buf = malloc(cb->nalloc += 2048);

    return cb->buf;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* Grow buffer if less than 1K free */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
char *
checksize(cb)
struct Ebuf *cb;
{
    if (cb->nalloc - cb->l  <  1024)
	return grow(cb);
    else
	return cb->buf;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* Check if string s1 ends with string s2 */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
char *ends(s1, s2)
char *s1, *s2;
/* If s1 ends with string s2, a pointer to the ending of s1 is returned;
 * else null
 */
{
    int l1, l2;
    l1 = strlen(s1);
    l2 = strlen(s2);
    if (l1 < l2)
	return NULL;
    else if (strcmp(s1+l1-l2, s2) == 0)
	return s1+l1-l2;
    else
	return NULL;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* Do fgets to get one logical line: join lines that are terminated
 * with backslash-newline.  Don't discard backslash or newline (so that
 * we can print the exact text, if desired).
 * The result is stored in "ebuf" and a pointer to the string
 * is returned.
 */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
char *
fillbuffer(fp, nl)
FILE *fp;
int *nl;	/* If non-null, number of lines read in is returned */
{
    char *s;
    ebuf.l = 0;		/* clear the extensible buffer */

    /* Collect lines until we have a non-zero buffer (which happens with
     * the first line, of course) and it isn't terminated "\\\n".
     */
    if (nl) *nl = 0;
    UNTIL(ebuf.l && !(s=ends(ebuf.buf, "\\\n"))) {
	if (!checksize(&ebuf)) {
	    /* Needed to, but couldn't increase the allocated space */
	    return NULL;
	}
	if (!fgets(ebuf.buf+ebuf.l, ebuf.nalloc - ebuf.l, fp))
	    return NULL;
	if (ebuf.l != 0 && !isspace(*(ebuf.buf+ebuf.l))) {
	    (void) fprintf(stderr,
		"%s: SUPER.TAB FORMAT ERROR -- Continued line not indented\n",
		prog);
	    (void) fprintf(stderr,
		"\tProblem text:\n%s\n",
		ebuf.buf);
	    (void) exit(1);
	}
	ebuf.l += strlen(ebuf.buf+ebuf.l);
	if (nl) (*nl)++;
    }
    return ebuf.buf;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* Look up command "usrcmd" in file "superfile".  Return path to execute
 * if approved, empty string if no action should be taken, NULL ptr
 * if error.
 */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
char *
approve(superfile, usrcmd)
char *superfile;	/* file of approved commands+users */
char *usrcmd;		/* command we're to check on.
			 * If command is one of:
			 *	"" | "-h" | "-?"
			 *	then a list of super commands is printed
			 *	instead of attempting to execute anything.
			 */
{
    int uid;
    struct passwd *usrpw;
    /* has to be static: holds the returned path used by the caller. */
    char *userlist, *command, *path, *tok;
    char *wd, **wdlist;
    int globbraces();
    char *word(), *re_comp(), *check_ugh();
    char *buf, *s;
    char *sep;
    int i, iwd, nl, line, givehelp, commandfound;
    int gid;
    int gethostname();
    char hostname[1024];
    void anchor();

    uid = getuid();
    usrpw = getpwuid(uid);
    if (!usrpw) {
	(void) fprintf(stderr,
		"%s - approve(): Couldn't get your password entry: ", prog);
	perror("");
	return NULL;
    }
    user = usrpw->pw_name;
    gid = usrpw->pw_gid;

    if (gethostname(hostname, sizeof(hostname)) == -1) {
	(void) fprintf(stderr, "%s: Couldn't get your hostname: ",
				prog);
	perror("");
	return NULL;
    }

    if ((fp = fopen(superfile, "r")) == NULL) {
	(void) fprintf(stderr, "%s: Couldn't open valid-users text file `%s': ",
				prog, superfile);
	perror("");
	return NULL;
    }

    sep = " \t\v\n";		/* How to split fields on input lines */

    /* Do we just give help or really match a command with this user? */
    givehelp = (usrcmd == NULL) ||
			strcmp(usrcmd, "-h")==0 || strcmp(usrcmd, "-?")==0;

    if (givehelp) {
	(void) fprintf(stderr, "%s version %s patchlevel %s\n",
				prog, Version, Patchlevel);
	(void) fprintf(stderr, "Use:\n\t%s command [args]\n\n", prog);
	(void) fprintf(stderr, "%-16s  %-24s  %s\n",
		"Command", "Full Path           ",
					"Valid User:Group@Host Patterns");
	(void) fprintf(stderr, "%-16s  %-24s  %s\n",
		"-------", "--------------------",
					"------------------------------");
    }

    commandfound = 0;
    for (line=1; buf=fillbuffer(fp, &nl); line += nl) {
	if (!*buf || *buf == '#' )		/* Strip comments */
	    continue;
	buf = strtok(buf, "#");			/* Strip more comments */
	if (re_comp("^[ \t\v\n]*$") != NULL) {	/* Skip whitespace */
	    paterror("bad whitespace pattern!", buf, line, nl, superfile);
	    return NULL;
	}
	if (re_exec(buf) == 1)
	    continue;

	if (isspace(*(ebuf.buf))) {
	    (void) fprintf(stderr,
	"%s: SUPER.TAB FORMAT ERROR -- Commandnames must begin in column 1\n",
	    prog);
	    (void) fprintf(stderr, "\tline %d, file %s\n", line, superfile);
	    (void) exit(1);
	}
						
	userlist = word(buf, " \t\v\n", 3);	/* Find the okusernames before
						 * carving up the line.
						 */
	command = strtok(buf, sep);		/* get the command... */
	path = strtok(NULL, sep);		/* ...and its full path */
	if (!command || !path)
	    continue;				/* Skip improper lines */

	if (givehelp) {
	    (void) fprintf(stderr,
		"%-16s  %-24s  %s",
		command, path, userlist);
	    continue;
	}

	if (strcmp(command, usrcmd) != 0)
	    continue;				/* Skip non-matching commands */

	commandfound++;

	/* Check our user against all valid user patterns */
	if (uid == 0)
	    return path;			/* root is always legit */

	for (wd = strtok(NULL, sep); wd; wd = strtok(NULL, sep)) {
	    if ((i=globbraces(wd, &wdlist)) != 0) {
		char errmsg[20];
		sprintf(errmsg, "Missing `%c'", i);
		paterror(errmsg, buf, line, nl, superfile);
	    } else {
		for (iwd=0; tok=wdlist[iwd]; iwd++) {
		    s = check_ugh(superfile, wd, tok, line, nl,
				user, gid, hostname);
#ifdef DEBUG
		    if (s == NULL) {
			(void) fprintf(stderr,
			    "MATCH:\n\tpath = %s\n\tpat=%s\n", path, wd);
			    return "";
		    } else {
			(void) fprintf(stderr,
			    "user=%s gid=%d hostname=%s; pat=%s; reason: %s\n",
			    user, gid,hostname,tok, s);
		    }
#else
		    if (s == NULL)
			return path;
#endif
		}
	    }
	}
    }
    if (givehelp)
	return "";
    else if (!commandfound)
	(void) fprintf(stderr, "%s: No such super command as `%s'.\n",
		prog, usrcmd);
    else
	(void) fprintf(stderr, "%s %s: Permission denied to user %s\n",
		prog, usrcmd, user);
    return NULL;
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* Check a single user/group/host string */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
char *
check_ugh(superfile, origtext, token, line, nl, user, gid, hostname)
char *superfile;	/* used just for reporting errors about this file */
char *origtext;		/* original text containing token -- for err messages */
char *token;		/* user/group/host to check */
int line;		/* line number in superfile, for reporting errors */
int nl;			/* number of lines (ie lines n..n+nl-1 are the present
			 * 	block)
			 */
char *user;		/* username to check */
int gid;		/* gid for user */
char *hostname;		/* host to check */
{
    char tok[1024];
    char chkbuf[1024];
    int re_exec(), ingroup();
    char *userpat, *grouppat, *hostpat;
    int i;

    /* Strip trailing backslash that may have preceded a newline -- it was
     * until now so that we can print, if necessary, the identical text
     * to that in the super.tab file.
     */
    (void) strcpy(tok, token);
    if ((i = strlen(tok)) && tok[i-1] == '\\') tok[i-1] = '\0';

    /* Split into user:group@host */
    hostpat = strchr(tok, '@');
    if (hostpat && !*(hostpat+1))
	return paterror("Missing hostname", origtext, line, nl, superfile);
    else if (hostpat)
	*hostpat++ = '\0';

#ifdef USE_NETGROUP
    if (hostpat && *hostpat == '+') {	/* Interpret as +netgroupname */
	if (*(hostpat+1) == '\0')
	    return paterror("Missing netgroupname",
					origtext, line, nl, superfile);
	if (!innetgr(hostpat+1, hostname, (char *) NULL, (char *) NULL))
	    return "Not in netgroup";	/* Not in this netgroup */
    }
#else
    (void) fprintf(stderr,
	"%s: hostnames may not begin with `+' since this super()\n", prog);
    (void) fputs("was compiled without ``#define USE_NETGROUP''\n", stderr);
    return 0;
#endif

    if (hostpat && *hostpat != '+') {	/* Interpret as a hostname pattern */
	anchor(hostpat, chkbuf);	/* Force all matches to be anchored */
	if (re_comp(chkbuf) != NULL)
	    return paterror("bad host pattern", origtext, line, nl, superfile);
	if (re_exec(hostname) != 1)
	    return "Doesn't match hostname";	/* Didn't match hostname */
    }

    grouppat = strchr(tok, ':');
    userpat = tok;
    if (grouppat && *(grouppat+1)) {	/* pat is "uuu:ggg or ":ggg" */
	if (tok == grouppat)
	    userpat = "^.*$";		/* pat is ":ggg" */
	grouppat++;
    } else {				/* pat is "uuu" or "uuu:" */
	if (grouppat)
	    *grouppat = '\0';		/* pat was "uuu:" */
	grouppat = "^.*$";
    }
    anchor(userpat, chkbuf);		/* Force all matches to be anchored */
    if (re_comp(chkbuf) != NULL)
	return paterror("bad user pattern", origtext, line, nl, superfile);
    if (re_exec(user) != 1)
	return "Doesn't match user";	/* Didn't match user */

    anchor(grouppat, chkbuf);
    i = ingroup(user, gid, chkbuf);
    if (i == -1)
	return paterror("bad group pattern", origtext, line, nl, superfile);
    else if (i != 1)
	return "Doesn't match group";
    
    return NULL;			/* Success! */
}
